-----------------------------------------------------------------------------
-- |
-- Module      :  Code
-- Project     :  HBC
-- Copyright   :  (c) Hal Daume III
-- License     :  Open
-- 
-- Maintainer  :  Hal Daume III <me@hal3.name>
-- Stability   :  stable
-- Portability :  portable
--
-- Abstract code generation
--
-----------------------------------------------------------------------------

module Code where

import Type
import Core(Const(..))
import qualified Core as C
import qualified Data.Map as M
import Data.Maybe
import Data.Char
import Data.List
import Debug.Trace 
import Control.Monad
import Control.Monad.State
import Util
import Marginalize
import Data.Generics hiding ((:*:))
import Data.Typeable


type Code = [Function]

type Id = String

data Function = 
     Function { 
       name       :: Id,
       returnVal  :: Maybe Expr,
       returnType :: Type,
       params     :: [(Id,Type)],
       tempVars   :: [(Id,Type)],  -- these needn't be created in body
       body       :: [Statement]
     } deriving (Eq, Ord, Typeable, Data)

data V = V Id [Expr] deriving (Eq, Ord, Typeable, Data)

data Statement = 
      Loop    { loopVar :: Id, loopBegin :: Expr, loopEnd :: Expr, loopBody :: [Statement] }
    | If      { condExp :: Expr, condThen :: [Statement], condElse :: [Statement]}
    | Assn    { assnVar :: V , assnVal :: Expr }
    | Incr    { incrVar :: V , incrVal :: Expr }
    | Call    { callFun :: Id, callRet :: [Expr], callArg :: [Expr] }  -- different languages handle return values differently
    | Comment { comment :: String }
    deriving (Eq, Ord, Typeable, Data)

data Expr =
      Var V
    | Con C.Const
--    | Ind Expr [Expr]
    | Fun Id [Expr]
    deriving (Eq, Ord, Typeable, Data)

type TmpVar = ((Id,Type),[Statement])

-- in order to translate sums and products, we need to create
-- auxiliary variables.  all of these generated statements
-- are guaranteed to be Loops.  the id/type included is the
-- new id created and its type (to be added to "tempVars")
codifyExpr :: TypeMap -> C.Expr -> (Expr, [TmpVar])
codifyExpr tm (C.Var (C.V id [])) = (Var (V id []), [])
codifyExpr tm (C.Var (C.V id sl)) = (Var (V id sl'), sts)
  where (sl', sts) = codifyMany (codifySExpr tm) sl
codifyExpr tm (C.Con c)           = (Con c, [])

codifyExpr tm (C.Func "sub" [a,b]) =
  case a' of
    Var (V a' ax') -> (Var (V a' (ax'++[b'])), stsA++stsB)
    Fun "+" [Fun "vec" (vecVal:_), Var (V a' ax')] -> (Fun "+" [vecVal, Var (V a' (ax'++[b']))], stsA++stsB)
    Fun ".+" [lhs, Var (V a' ax')] -> (Fun ".+" [lhs, Var (V a' (ax'++[b']))], stsA++stsB)
    e -> error ("Code.codifyExpr: cannot codify " ++ show e)
  where (a', stsA) = codifyExpr tm a
        (b', stsB) = codifyExpr tm b

{-
codifyExpr tm (C.Func "sub" [a,b]) = (Var (V a' (ax'++[b'])), stsA++stsB)
  where (Var (V a' ax'), stsA) = codifyExpr tm a
        (b', stsB) = codifyExpr tm b
-}

codifyExpr tm (C.Func f el)       = (Fun f el', sts)
  where (el', sts) = codifyMany (codifyExpr tm) el
codifyExpr tm sp@(C.SP b r e var)
  | null var  = error ("Code.codifyExpr: got unnamed SP in: " ++ show sp)
  | otherwise = (Var (V var []), ((var,typ),st):sts0'++sts1)
  where
    typ = case M.lookup var tm of { Nothing -> error ("Code.codifyExpr: could not find type of sum/product " ++ show var ++ " in: " ++ show sp) ; Just t -> t }
    (e',sts0) = codifyExpr tm e
    (sts0',states0) = splitStatements sts0
--    r' = fixRange tm var e r
    (st,sts1) = sumProductStatement tm var typ b r e' states0

{-

fixRange tm sp e (C.Range v0 (C.Var (C.V "?" [])) (C.Var (C.V "?" []))) 
  | otherwise = error ("Code.fixRange: " ++ show (v0,varEq,e,sp,tm))
  where
    varEq = nub $ findEquality e

    findEquality (C.Func "=" [C.Var (C.V v []), e']) | v0 == v = [e']
    findEquality (C.Func "=" [e', C.Var (C.V v [])]) | v0 == v = [e']
    findEquality (C.Func _ el) = concatMap findEquality el
    findEquality (C.SP _ _ e _) = findEquality e
    findEquality _ = []
fixRange _ _ _ r = r
-}



codifySExpr tm (C.Id id Nothing) = (Var (V id []), [])
codifySExpr tm (C.Id id (Just c)) = (Fun "+" [Var (V id []), Con c], [])
codifySExpr tm (C.Cn c) = (Con c, [])

sumProductStatement tm var typ b (C.Range id lo hi) e rest = -- (C.SP b (C.Range id lo hi) e) =
  ([Assn var0 (if b then one else zero),
    Loop id lo' hi' $
      rest ++
      [if b 
       then Assn var0 (Fun "*" [Var var0, e])
       else Incr var0 e]
   ],
   ((id,tInt),[]):stsLo ++ stsHi)
  where
    var0 = V var []

    (lo', stsLo) = codifyExpr tm lo
    (hi', stsHi) = codifyExpr tm hi

    (zero,zeroS) = mkVec (indices typ) $ Con (if baseType typ == T0Real then C.R 0 else C.I 0)
    (one ,oneS ) = mkVec (indices typ) $ Con (if baseType typ == T0Real then C.R 1 else C.I 1)

    mkVec [] e = (e, [])
    mkVec l  e = (Fun "vec" (e : l'), sts)
      where 
        (l', sts) = codifyMany (codifyExpr tm) $ concatMap rangeToExpr l

rangeToExpr (C.Range _ lo hi) = [lo,hi]
    

--    (e' , stsE ) = codifyExpr tm e

{-
codifyLDF tm (_, dist, rl) = (e, sts ++ sts')
  where
    (dist', sts ) = codifyDistLDF tm dist
    (rl',   tm' ) = applyRanges tm rl dist'
    (e    , sts') = codifyExpr    tm rl'

    applyRanges tm []     e = (e , tm)
    applyRanges tm (r:rs) e = (e', M.insert (C.spVarName sp) tReal tm')
      where
        (e', tm') = applyRanges tm rs e
        sp        = C.SP False r e'
-}

codifyVar tm (C.V id []) = Var (V id [])
codifyVar tm (C.V id sl) = Var (V id (map (fst . codifySExpr tm) sl))

codifyDistLDF tm (C.Dist var "IG" el) = (Fun "ldf_Gam" ((Fun "/" [Con (R 1), codifyVar tm var]) : el'), sts)
  where (el', sts) = codifyMany (codifyExpr tm) el
codifyDistLDF tm (C.Dist var "Delta" el) = (Con (C.R 0), [])
codifyDistLDF tm (C.Dist var dist el) = (Fun ("ldf_" ++ dist) (codifyVar tm var : el'), sts)
  where (el', sts) = codifyMany (codifyExpr tm) el
codifyDistLDF tm ((C.:*:) a b) = (Fun "+" [a',b'], stsA ++ stsB)
  where (a', stsA) = codifyDistLDF tm a
        (b', stsB) = codifyDistLDF tm b
codifyDistLDF tm ((C.:^:) a e) = (Fun "*" [e',a'], stsA ++ stsE)
  where (a', stsA) = codifyDistLDF tm a
        (e', stsE) = codifyExpr tm e
codifyDistLDF tm prod@(C.Prod r d var) 
  | null var  = error ("Code.codifyDistLDF: unnamed product variable found in: " ++ show prod)
  | otherwise = (Var (V var []), ((var,typ),st):sts0'++sts1)
  where
    typ = tReal
    tm' = M.insert var typ tm
    (d',sts0) = codifyDistLDF tm' d
    (sts0',states0) = splitStatements sts0
    (st,sts1) = sumProductStatement tm var typ False r d' states0

codifyLDF tm (nn, (_, dist, rl)) = codifyDistLDF tm $ applyRanges 0 rl dist
  where
    applyRanges mm []     d = d
    applyRanges mm (r@(C.Range v _ _):rs) d =
      C.Prod r' (applyRanges (mm+1) rs d) ("ldfP" ++ show nn ++ "_" ++ show mm)
      where 
        off = filter (/=0) $ C.getOffsets v d
        r'  = case off of { [] -> r ; os -> C.applyRangeOffset (minimum os) (maximum os) r }

splitStatements sl = ([(i,[]) | (i,_) <- sl], concatMap snd sl)

codifyMany :: (a -> (b, [TmpVar])) -> [a] -> ([b], [TmpVar])
codifyMany f [] = ([], [])
codifyMany f (x:xs) =
  let (y ,sts ) = f x
      (ys,sts') = codifyMany f xs
  in  (y:ys, sts++sts')


------------------------------------------------------------------------------------------------------------------------------

makeLikFunction :: C.Core -> [Id] -> TypeMap -> Function
makeLikFunction (C.Core dl) globals tm =
  Function { name       = "compute_log_posterior"
           , returnVal  = Just ret
           , returnType = tReal
           , params     = [(v, fromMaybe (error $ "makeLikFunction: Nothing on type lookup for " ++ show v) $ M.lookup v tm) | v <- globals]
           , tempVars   = temps
           , body       = concatMap snd sts }
  where
    (el, sts) = codifyMany (codifyLDF tm) (zip [0..] dl)
    temps     = map fst sts
    ret       = foldr1 (\a b -> Fun "+" [a,b]) el

------------------------------------------------------------------------------------------------------------------------------

makeSampleFunction :: C.Decl -> TypeMap -> [Update] -> Bool -> Function
makeSampleFunction decl@(v0@(C.V sampVar varIdx), dist, rl) tm updates maximizeIt =
  Function { name       = "resample_" ++ sampVar
           , returnVal  = Just (Var (V sampVar []))
           , returnType = typ0
           , params     = nub [(v, fromMaybe (error $ "makeSampleFunction: Nothing on type lookup for " ++ show v) $ M.lookup v tm) | v <- myGlobals ++ newGlob, not ("tmp" `isPrefixOf` v), not (v `elem` rangeVars)]
           , tempVars   = temps'
           , body       = body'  }
  where
    rangeVars = [id | C.Range id _ _ <- rl]
    typ0@(Type t0 t0i) = fromMaybe (error $ "makeLikFunction: Nothing on type lookup for sampVar " ++ show sampVar) $ M.lookup sampVar tm
    myGlobals = sort $ nub $ C.getDeclGlobals decl

    {- we structure the sampling function as follows.  "sampVar" will always come
       in as a parameter, because it's considered one of my globals.  we will
       always *overwrite* this variable *and* set it as the return val (the return
       val will always be a simple Var).  this means that if we're generating code
       in a pass-by-reference language, and the variable is passed by reference,
       then we can just return void.  OTOH, if it's passed by value, we'll need
       to return it.  overwriting will also give us speedups in matlab, where
       this will not result in the creation of an temporary copy
    -}

    (body , temps , newGlob) = makeSampleBody v0 tm dist maximizeIt
    ((updBef, updAft), updTmp) = generateMarginalUpdates tm updates sampVar varIdx
    (body', temps') = foldr (addLoopRange tm) (updBef ++ body ++ updAft, temps ++ map fst updTmp) rl


addLoopRange :: TypeMap -> C.Range -> ([Statement], [(Id,Type)]) -> ([Statement], [(Id,Type)])
addLoopRange tm (C.Range rid lo hi) (b,t) = 
  ( (concatMap snd (stsLo++stsHi)) ++ [Loop rid lo' hi' b]
  , (rid,tInt) : t ++ map fst (stsLo++stsHi))
  where
    (lo', stsLo) = codifyExpr tm lo
    (hi', stsHi) = codifyExpr tm hi
--    off = filter (/=0) $ getOffsets rid 

------------------------------------------------------------------------------------------------------------------------------

generateMarginalUpdates :: TypeMap -> [Update] -> Id -> [C.SExpr] -> (([Statement],[Statement]), [TmpVar])
generateMarginalUpdates tm updates id0 rl =
  case filter ((==id0).indexVariable) updates of
    [] -> (([],[]), [])
    up -> 
      let (stmts,idts) = unzip $ map (generateMarginalUpdate tm id0 rl) up
          (befS, aftS) = unzip stmts
      in  ((concat befS, concat $ reverse aftS), nub $ concat idts)

generateMarginalUpdate :: TypeMap -> Id -> [C.SExpr] -> Update -> (([Statement],[Statement]), [TmpVar])
generateMarginalUpdate tm id0 rl (Update _ isIncr _ varRem varAdd) = ((bef, aft), sts)
  where
    (bef,befS) = mkS $ varRem rl -- (map C.rangeIds rl)
    (aft,aftS) = mkS $ varAdd rl -- (map C.rangeIds rl)
    sts = befS ++ aftS

    mkS (ranges, v0, expr) =
      let (e0, tmp)   = codifyExpr tm expr
          (Var v, []) = codifyExpr tm (C.Var v0)
          s0          = if isIncr then Incr v e0 else Assn v e0
          (s, it)     = foldr (addLoopRange tm) ([s0], []) ranges
      in  (s, map (\x -> (x,[])) it ++ tmp)
    

------------------------------------------------------------------------------------------------------------------------------

makeDumpFunction :: C.Core -> C.Decl -> TypeMap -> Function
makeDumpFunction core@(C.Core decls) decl@(v0@(C.V sampVar varIdx), dist, rl) tm =
  Function { name       = "dump_" ++ sampVar
           , returnVal  = Nothing
           , returnType = tInt
           , params     = nub ([(v, fromMaybe (error $ "makeDumpFunction: Nothing on type lookup for " ++ show v) $ M.lookup v tm) | v <- myGlobals, not ("tmp" `isPrefixOf` v), not (v `elem` lhsVars), not (v `elem` rangeVars)])
           , tempVars   = temps
           , body       = body'  }
  where
    lhsVars = [id | (C.V id _,_,_) <- decls, id /= sampVar]
    rangeVars = [id | C.Range id _ _ <- rl]

    typ0@(Type t0 t0i) = fromMaybe (error $ "makeDumpFunction: Nothing on type lookup for sampVar: " ++ show sampVar) $ M.lookup sampVar tm
    myGlobals = sort $ nub $ (sampVar : C.getDeclGlobals decl)

    (Var (V v0' _),[]) = codifyExpr tm (C.Var v0)

    (body,  temps )  = makeDumpBody 1 (Var (V v0' [])) tm typ0
    body' = (Call "printf_string" [] [Var (V sampVar [])]) : body ++ [Call "printf_newline" [] []]

    zero = case t0 of { T0Real -> Con (C.R 0) ; _ -> Con (C.I 0) }

makeDumpBody _ v0 tm (Type T0Int  []) = ([Call "printf_int"  [] [v0]], [])
makeDumpBody _ v0 tm (Type T0Real []) = ([Call "printf_real" [] [v0]], [])
makeDumpBody z (Var (V v0 v0i)) tm (Type t0 ((C.Range _ lo hi):t0i)) =
  ([Loop tIdx lo' hi' body''], map fst loV ++ map fst hiV ++ vars')
  where
    tIdx = "dvv_loop_var_" ++ show z
    (lo', loV) = codifyExpr tm lo
    (hi', hiV) = codifyExpr tm hi
    (body', vars') = makeDumpBody (z+1) (Var (V v0 (v0i ++ [(Var (V tIdx []))]))) tm (Type t0 t0i)
    body'' = body' ++ [Call "printf_sep" [] [Con (C.I (genericLength t0i))]]

    


------------------------------------------------------------------------------------------------------------------------------

makeInitializeFunction :: C.Core -> C.Decl -> TypeMap -> Function
makeInitializeFunction core@(C.Core decls) decl@(v0@(C.V sampVar varIdx), dist, rl) tm
--  | trace (show (myGlobals,newGlob,lhsVars)) False = undefined
  | otherwise =
  Function { name       = "initialize_" ++ sampVar
           , returnVal  = Just varNew
           , returnType = typ0
           , params     = nub ([(v, fromMaybe (error $ "makeInitializeFunction: Nothing on type lookup for " ++ show v) $ M.lookup v tm) | v <- myGlobals ++ newGlob, not ("tmp" `isPrefixOf` v), not (v `elem` lhsVars), not (v `elem` rangeVars)])
           , tempVars   = (sampVar, typ0) : temps''
           , body       = body''  }
  where
    varNew  = Var (V sampVar [])

    lhsVars = [id | (C.V id _,_,_) <- decls]
    rangeVars = [id | C.Range id _ _ <- rl]

    typ0@(Type t0 t0i) = fromMaybe (error $ "makeInitializeFunction: Nothing on type lookup for sampVar: " ++ show sampVar) $ M.lookup sampVar tm
    myGlobals = sort $ nub $ C.getDeclGlobals decl

    (v0',[]) = codifyExpr tm (C.Var v0)

    (body,  temps , newGlob)  = makeInitializeBody v0' tm dist
    (body', temps')  = foldr (addLoopRange tm) (body, temps) rl
    (body'',temps'') = 
      if null t0i then (body',temps') else
        let (l',sts) = codifyMany (codifyExpr tm) $ concatMap rangeToExpr t0i
        in  ((Assn (V sampVar []) (Fun "vec" (zero:l'))) : body', map fst sts ++ temps')

    zero = case t0 of { T0Real -> Con (C.R 0) ; _ -> Con (C.I 0) }

makeMarginalInitialization :: TypeMap -> Code -> Id -> Function
makeMarginalInitialization tm code id =
  case find ((==("resample_" ++ id)).name) code of
    Nothing -> error ("Code.makeMarginalInitialization: cannot find resample function for " ++ id)
    Just f  ->
      let Type t0 t0i = fromMaybe (error $ "makeMarginalInitialization: Nothing on type lookup for " ++ show id) $ M.lookup id tm
          zero = case t0 of { T0Real -> Con (C.R 0) ; _ -> Con (C.I 0) }
          (l',[]) = codifyMany (codifyExpr tm) $ concatMap rangeToExpr t0i
      in  Function { name = "initialize_" ++ id
                   , returnVal  = returnVal  f
                   , returnType = returnType f
                   , params     = filter ((/=id).fst) $ params f
                   , tempVars   = [(id, Type t0 t0i)]
                   , body       = [
                                    Assn (V id []) (Fun "vec" (zero:l'))
                                  , Call (name f) [] (map (\ (id,_) -> Var (V id [])) $ params f)
                                  ] }

------------------------------------------------------------------------------------------------------------------------------

makeInitializeBody :: Expr -> TypeMap -> C.Dist -> ([Statement], [(Id,Type)], [Id]) -- final is new GLOBALS for multsym
makeInitializeBody v0 tm (C.Dist var dist el) =
  case dist of
    "DirSym" ->
      let (el',tv) = codifyExpr tm (el !! 1)
      in  (concatMap snd tv ++ [Call "sample_DirSym" [v0] [one, el']], map fst tv, [])

    "Dir"    ->
      let (el',tv) = codifyExpr tm (el !! 1)
      in  (concatMap snd tv ++ [Call "sample_DirSym" [v0] [one, el']], map fst tv, [])

    "Bin"    -> ([Call "sample_Bin"   [v0] [half]], [], [])
    "Poi"    -> ([Call "sample_Poi"   [v0] [one]], [], [])
    "Gam"    -> ([Call "sample_Gam"   [v0] [one , one]], [], [])
    "IG"     -> ([Call "sample_Gam"   [v0] [one , one], Assn v0var (Fun "/" [one, v0])], [], [])
    "Bet"    -> ([Call "sample_Bet"   [v0] [one , one]], [], [])
    "Nor"    -> ([Call "sample_Nor"   [v0] [zero, one]], [], [])
    "NorMV"  -> (concatMap snd sts ++ [Call "sample_NorMV" [v0] [zerV, one]], map fst sts, [])
      where 
        (zerV,sts) = 
          case typEl !! 0 of
            Just (Type _ dims) -> 
              let (l',sts) = codifyMany (codifyExpr tm) $ concatMap rangeToExpr dims
              in  (Fun "vec" (zero:l'), sts)
            t -> error ("Code.makeInitializeBody: don't know how to deal with type " ++ show t ++ " in NorMV")

    "Mult"   ->
      case typEl !! 0 of
        Just (Type _ [C.Range _ (C.Con (C.I 1)) hi]) ->
          let (hi',tv) = codifyExpr tm hi
          in  (concatMap snd tv ++ [Call "sample_MultSym" [v0] [Con (C.I 1),hi']], map fst tv, C.getExprGlobals [] hi)
        _ -> error ("cannot infer type of Mult in " ++ show dist ++ ", typEl=" ++ show typEl)

    "MultSym" ->
      let (el', tv) = codifyMany (codifyExpr tm) el
      in  (concatMap snd tv ++ [Call "sample_MultSym" [v0] el'], map fst tv, [])


--          in  (concatMap snd tv ++ [Call "sample_Mult" [v0] [Fun "vec" [Fun "/" [one, hi'], Con (C.I 1), hi']]], map fst tv)

    "Delta" ->
      case typEl !! 0 of
        Just (Type T0Real [C.Range _ (C.Con (C.I 1)) hi]) ->
          let (hi',tv) = codifyExpr tm hi
          in  (concatMap snd tv ++ [Call "copy_vector" [v0] [Con (C.I 1),hi']], map fst tv, C.getExprGlobals [] hi)

        _ -> error ("Code.makeInitializeBody: cannot infer type of " ++ show (el !! 0) ++ "; found " ++ show (typEl !! 0))
        

  where typEl = map (inferType tm) el
        half  = Con (C.R 0.5)
        one   = Con (C.R 1)
        zero  = Con (C.R 0)
        Var v0var = v0

makeInitializeBody _ _ d = error ("Code.makeInitializeBody: expecting Dist, maybe you forgot to pass the prior?  Got " ++ show d)

------------------------------------------------------------------------------------------------------------------------------

makeSampleBody :: C.Var -> TypeMap -> C.Dist -> Bool -> ([Statement], [(Id,Type)], [Id])
makeSampleBody _ tm d@(C.Dist var dist el) maximizeIt = -- this is the easy case: conjugacy has rescued us
  (comm ++ sts ++ sampleCall, map fst tmpVars, [])
  where
    (el', tmpVars) = codifyMany (codifyExpr tm) el
    sts = concatMap snd tmpVars
    v0@(Var v0v)  = fst $ codifyExpr tm (C.Var var)
    sampleCall = 
      case dist of
        "IG" -> [Call (if maximizeIt then "mode_Gam" else "sample_Gam") [v0] el', Assn v0v (Fun "/" [Con (C.R 1), v0])]
        _    -> [Call ((if maximizeIt then "mode_" else "sample_")++dist) [v0] el']
    comm = [Comment ("Implements direct sampling from the following distribution:")
           ,Comment ("  " ++ show d)]

makeSampleBody v0@(C.V sampVar varIdx) tm dist maximizeIt
  | t0 == T0Int && length t0i == length varIdx &&     -- check if it's discrete
    isJust varOcc && dist0 == "Mult" &&               -- and appears in a multinomial
    isJust typTh && thLoIsOne                         -- make sure we can figure out the dimensions
    = (comm ++ body, 
        (tPost, Type T0Real rtheta) :
        (tIdx , tInt) :
        map fst stsHi ++
        map fst stsD,
       C.getExprGlobals [] hi)
            
  where
    Type t0 t0i = fromMaybe (error $ "makeSampleBody (Mult): Nothing on type lookup for sampVar " ++ show sampVar) $ M.lookup sampVar tm
    varOcc      = C.findVariableOccurance v0 dist
    Just (C.Dist _ dist0 (theta:_)) = varOcc
    typTh       = inferType tm theta
    Just (Type _ rtheta@[C.Range _ lo hi]) = typTh
    thLoIsOne = lo == C.Con (C.I 1)
    {- to create the body, we need a temporary R[hi] to store the log
       probabilities.  we also need a temporary Z to iterate over these
       positions.  we fill in each position with a sum of ldfs.  then,
       we normalize and sample a multinomial -}
    tPost = makeTmpVar "post" v0
    tIdx  = makeTmpVar "idx"  v0

    (hi',stsHi0)    = codifyExpr tm hi
    (stsHi,stateHi) = splitStatements stsHi0

    Just dist'      = C.replaceVar v0 tIdx dist

    (df, stsD0)     = codifyDistLDF tm dist'
    (stsD,stateD)   = splitStatements stsD0

    body =
        stateHi ++ 
        [ Assn (V tPost []) (Fun "vec" [Con (C.R 0), Con (C.I 1), hi'])
        , Loop tIdx (Con (C.I 1)) hi'
               (stateD ++ 
                [Assn (V tPost [Var (V tIdx [])]) df]
               )
        , Call "normalizeLog" [] [Var (V tPost [])]
        , Call (if maximizeIt then "mode_Mult" else "sample_Mult")  [fst $ codifyExpr tm (C.Var v0)] [Var (V tPost [])]
        ]
    comm = [Comment ("Implements multinomial sampling from the following distribution:")
           ,Comment ("  " ++ show dist)]

makeSampleBody v0@(C.V sampVar varIdx) tm dist maximizeIt
  | t0 == T0Int && length t0i == length varIdx &&     -- check if it's discrete
    isJust varOcc && dist0 == "Bin"                   -- and appears in a binomial
    = (comm ++ body, 
        (tPost, Type T0Real [C.Range "" (C.Con (C.I 1)) (C.Con (C.I 2))]) :
        (tIdx , tInt) :
        map fst stsD,
       C.getExprGlobals [] (C.Con (C.I 2)))
            
  where
    Type t0 t0i = fromMaybe (error $ "makeSampleBody (Bin): Nothing on type lookup for sampVar " ++ show sampVar) $ M.lookup sampVar tm
    varOcc      = C.findVariableOccurance v0 dist
    Just (C.Dist _ dist0 _) = 
      case varOcc of
        Just (C.Dist _ dist0 _) -> varOcc
        _ -> error ("Code.makeSampleBody (Bin): varOcc = " ++ show varOcc)

    {- to create the body, we need a temporary R[hi] to store the log
       probabilities.  we also need a temporary Z to iterate over these
       positions.  we fill in each position with a sum of ldfs.  then,
       we normalize and sample a multinomial -}
    tPost = makeTmpVar "post" v0
    tIdx  = makeTmpVar "idx"  v0

--    (hi',stsHi0)    = codifyExpr tm hi
--    (stsHi,stateHi) = splitStatements stsHi0

    Just dist'      = C.replaceVar v0 tIdx dist

    (df, stsD0)     = codifyDistLDF tm dist'
    (stsD,stateD)   = splitStatements stsD0

    body =
--        stateHi ++ 
        [ Assn (V tPost []) (Fun "vec" [Con (C.R 0), Con (C.I 1), Con (C.I 2)])
        , Loop tIdx (Con (C.I 1)) (Con (C.I 2))
               (stateD ++ 
                [Assn (V tPost [Var (V tIdx [])]) df]
               )
        , Call "normalizeLog" [] [Var (V tPost [])]
        , Call (if maximizeIt then "mode_Mult" else "sample_Mult")  [fst $ codifyExpr tm (C.Var v0)] [Var (V tPost [])]
        ]
    comm = [Comment ("Implements binomial sampling from the following distribution:")
           ,Comment ("  " ++ show dist)]

makeSampleBody _ _ d _ = 
    ([Comment "Code.makeSampleBody: I don't know how to sample in the following distribution; perhaps conjugacy application failed?",
      Comment ("  " ++ show d),
      Comment "Temporary solution: do nothing!"],
     [],[])


makeTmpVar str (C.V v l) = "tmp_" ++ str ++ "_" ++ v ++ "_" ++ show (length l)

------------------------------------------------------------------------------------------------------------------------------

-- removeID: remove occurances of the ID vector when possible
removeID :: Code -> Code
removeID = map (\f -> f { body = map  removeIDS $ body f })

removeIDS (Loop i lo hi b) = Loop i lo hi (map removeIDS b)
removeIDS (If c t e) = If c (map removeIDS t) (map removeIDS e)
removeIDS (Assn (V v sl) e) 
  | Just (val,idx) <- conditionalID e = Assn (V v (sl++[idx])) val
  | otherwise                         = Assn (V v sl) e
removeIDS (Incr (V v sl) e)
  | Just (val,idx) <- conditionalID e = Incr (V v (sl++[idx])) val
  | otherwise                         = Incr (V v sl) e
removeIDS x = x

conditionalID :: Expr -> Maybe (Expr, Expr)
conditionalID (Fun "ID" [idx,dim]) = Just (Con (C.I 1), idx)
conditionalID (Fun ".*" [a,b]) =
  case conditionalID b of
    Just (val, idx) -> Just (Fun "*" [val,a], idx)
    Nothing -> Nothing
conditionalID (Fun "*" [a,b]) =
  case conditionalID b of
    Just (val, idx) -> Just (Fun "*" [Fun "sub" [a,idx],val], idx)
    Nothing -> Nothing
conditionalID (Fun fn [c@(Con _), e])
  | fn `elem` ["-",".-"] = 
    case conditionalID e of
      Nothing -> Nothing
      Just (val, idx) -> Just (Fun fn [c, val], idx)
conditionalID _ = Nothing

------------------------------------------------------------------------------------------------------------------------------

removeSub :: Code -> Code
removeSub = map removeSubF

removeSubF f = f { returnVal = liftM removeSubE $ returnVal f
                 , body      = map   removeSubS $ body f      }

removeSubS (Loop id lo hi s) = Loop id (removeSubE lo) (removeSubE hi) (map removeSubS s)
removeSubS (If c t e) = If (removeSubE c) (map removeSubS t) (map removeSubS e)
removeSubS (Assn v e) = Assn (removeSubV v) (removeSubE e)
removeSubS (Incr v e) = Incr (removeSubV v) (removeSubE e)
removeSubS (Call f r a) = Call f (map removeSubE r) (map removeSubE a)
removeSubS (Comment c) = Comment c

removeSubV (V id e) = V id $ map removeSubE e

removeSubE (Fun "sub" ((Var (V id e0)):el)) = Var $ V id $ map removeSubE (e0++el)
removeSubE (Var v) = Var $ removeSubV v
removeSubE (Con c) = Con c
removeSubE (Fun i e) = Fun i $ map removeSubE e

------------------------------------------------------------------------------------------------------------------------------

-- giveNewVarNames :: Code -> Code
giveNewVarNames map0 m = evalState (mapM giveF m) (map0, 0)

giveF f = do
  mapM (fixName.fst) (params f)
  rv <- case returnVal f of { Just e -> giveE e >>= return . Just ; _ -> return Nothing }
  tv <- mapM  giveIT $ tempVars  f
  b  <- mapM  giveS  $ body      f
  mapM (unfixName.fst) (params f)
  return (f { returnVal = rv , tempVars = tv , body = b })

fixName   n = get >>= \ (m,i) -> put (M.insert n n m, i)
unfixName n = get >>= \ (m,i) -> put (M.delete n   m, i)

giveIT (i,t) = do
  i' <- getName i
  return (i',t)

giveS (Loop v lo hi b) = do
  v'  <- getName v
  lo' <- giveE   lo
  hi' <- giveE   hi
  b'  <- mapM giveS b
  return (Loop v' lo' hi' b')
giveS (If c t e) = do
  c' <- giveE c
  t' <- mapM giveS t
  e' <- mapM giveS e
  return (If c' t' e')
giveS (Assn v e) = do
  v' <- giveV v
  e' <- giveE e
  return (Assn v' e')
giveS (Incr v e) = do
  v' <- giveV v
  e' <- giveE e
  return (Incr v' e')
giveS (Call f r a) = do
  r' <- mapM giveE r
  a' <- mapM giveE a
  return (Call f r' a')
giveS (Comment c) = return (Comment c)

giveV (V i e) = do
  i' <- getName i
  e' <- mapM giveE e
  return (V i' e')

giveE (Var v) = giveV v >>= return . Var
giveE (Con c) = return (Con c)
giveE (Fun i e) = mapM giveE e >>= return . Fun i

getName n = get >>= \ (m,i) ->
  case M.lookup n m of
    Just n' -> return n'
    Nothing -> do
      let (n',i') = makeName m i
      let m' = M.insert n n' m
      put (m',i')
      return n'
  where
    makeName m i 
      | n `M.member` m = makeName m (i+1)
      | otherwise      = (n, i+1)
      where n = makeName0 i
          
    makeName0 i
      | i < 26    = [chr (i+97)]
      | otherwise = (chr ((i `mod` 26)+97)) : makeName0 (i `div` 26)

------------------------------------------------------------------------------------------------------------------------------

codeIdsF f = concatMap codeIdsE (maybeToList $ returnVal f) ++ concatMap codeIdsS (body f)

codeIdsS (Loop id lo hi b) = id : codeIdsE lo ++ codeIdsE hi ++ concatMap codeIdsS b
codeIdsS (If c t e) = codeIdsE c ++ concatMap codeIdsS (t++e)
codeIdsS (Assn v e) = codeIdsV v ++ codeIdsE e
codeIdsS (Incr v e) = codeIdsV v ++ codeIdsE e
codeIdsS (Call _ r a) = concatMap codeIdsE (r ++ a)
codeIdsS (Comment _) = []

codeIdsE (Var v) = codeIdsV v
codeIdsE (Con _) = []
codeIdsE (Fun _ e) = concatMap codeIdsE e

codeIdsV (V i e) = i : concatMap codeIdsE e

------------------------------------------------------------------------------------------------------------------------------

instance Show Function where
  showsPrec _ f =
    showString "Function " . showString (name f) . showChar '(' . showsParams (params f) . showChar ')' . showChar '\n' .
    (foldr (\x a -> showsTempVar x . a) id (tempVars f)) .
    (foldr (\x a -> showsPrec 1 x . a) id (body f)) .
    (case returnVal f of
       Nothing -> id
       Just e  -> showString "  Return " . showsPrec 0 e . showChar '\n') 
    where
      showsParams [] = id
      showsParams [(id,t)]    = showString id . showString " : " . shows t
      showsParams ((id,t):xs) = showString id . showString " : " . shows t . showString ", " . showsParams xs

      showsTempVar (id,t) = showString "  temporary " . showString id . showString " : " . shows t . showChar '\n'

instance Show Statement where
  showsPrec i s = showString (replicate (i*2) ' ') . showSt s
    where
      showSt (Loop v lo hi b) = 
        showString "For " . showString v . showString " = " . shows lo . showString " to " . shows hi . showString " do \n" . 
        foldr (\x a -> showsPrec (i+1) x . a) id b
      showSt (If cond thenS elseS) =
        showString "If " . shows cond . showString " Then\n" .
        foldr (\x a -> showsPrec (i+1) x . a) id thenS .
        (if null elseS then id else
          showString (replicate (i*2) ' ') . showString "Else \t {-- " . shows cond . showString " --}\n" .
          foldr (\x a -> showsPrec (i+1) x . a) id elseS)
      showSt (Assn v e) = shows v . showString " := " . shows e . showChar '\n'
      showSt (Incr v e) = shows v . showString " += " . shows e . showChar '\n'
      showSt (Call f [] e) = showString f . showChar '(' . showsParams e . showChar ')' . showChar '\n'
      showSt (Call f rl e) = showChar '[' . showsParams rl . showString "] <- " . showString f . showChar '(' . showsParams e . showChar ')' . showChar '\n'
      showSt (Comment s)   = showString " {-- " . showString s . showString " --}\n"

      showsParams []     = id
      showsParams [e]    = shows e
      showsParams (e:es) = shows e . showString ", " . showsParams es

instance Show V where
  showsPrec _ (V v []) = showString v
  showsPrec _ (V v el) = showString v . showString "_{" . showsParams el . showChar '}'
    where
      showsParams []     = id
      showsParams [e]    = shows e
      showsParams (e:es) = shows e . showString ", " . showsParams es

instance Show Expr where
  showsPrec _ (Var v) = shows v
  showsPrec _ (Con c) = shows c
--  showsPrec _ (Ind e []) = shows e
--  showsPrec _ (Ind e sl) = shows e . showString "_{" . showsParams sl . showChar '}'
  showsPrec _ (Fun f el) 
    | length el == 2 && not (any isAlphaNum f) = shows (el!!0) . showString f . shows (el!!1)
    | otherwise = showString f . showString "(" . showsParams el . showChar ')'
    where
      showsParams []     = id
      showsParams [e]    = shows e
      showsParams (e:es) = shows e . showString ", " . showsParams es

inferCodeType :: TypeMap -> Expr -> Type
inferCodeType tm (Var (V v sl)) =
  case M.lookup v tm of
    Nothing -> if '*' == head v && '@' `elem` v then tInt else error ("Code.inferCodeType: cannot find type of: " ++ v)
    Just (Type t0 t0i) -> 
        if length sl > length t0i
        then error ("Code.inferCodeType: too many subscripts on: " ++ v)
        else Type t0 (drop (length sl) t0i)
inferCodeType tm (Con (C.I _)) = tInt
inferCodeType tm (Con (C.R _)) = tReal
inferCodeType tm (Fun f el) = fromJust $ getFuncRange tm f (map (Just . inferCodeType tm) el) (Just $ map codeE2coreE el)

codeE2coreE (Var (V id el)) = C.Var (C.V id (map (fromJust . C.exprToSExpr . codeE2coreE) el))
codeE2coreE (Con c) = C.Con c
codeE2coreE (Fun i el) = C.Func i (map codeE2coreE el)

liftVecF :: TypeMap -> Function -> Function
liftVecF tm f = f { body     = body'' ++ body' 
                  , tempVars = tempVars f ++ newTemps }
  where
    (body', newVars, _) = liftMany liftVecS tm 0 $ body f
    body'' = mapMaybe thr3 newVars
    newTemps = [(id,t) | (Var (V id []), t, _) <- newVars]
    

liftMany f tm nn [] = ([], [], nn)
liftMany f tm nn (x:xs) = 
  let (y,  b,  nn1) = f tm nn x
      (ys, bs, nn2) = liftMany f tm nn1 xs
  in  (y:ys, b++bs, nn2)

liftVecS tm nn (Loop id lo hi b) = 
  let (lo', z0, nn0) = liftVecE tm nn  lo
      (hi', z1, nn1) = liftVecE tm nn0 hi
      (b' , z2, nn2) = liftMany liftVecS tm nn1 b 
  in  (Loop id lo' hi' b', z0++z1++z2, nn2)
liftVecS tm nn (If c t e) = 
  let (c', z0, nn0) = liftVecE tm nn c
      (t', z1, nn1) = liftMany liftVecS tm nn0 t
      (e', z2, nn2) = liftMany liftVecS tm nn1 e
  in  (If c' t' e', z0++z1++z2, nn2)
liftVecS tm nn (Assn v (Fun "vec" el)) = liftVecS tm nn (Call "vec" [] ((Var v):el))
liftVecS tm nn (Assn v (Fun "ID"  el)) = liftVecS tm nn (Call "ID"  [] ((Var v):el))
liftVecS tm nn (Assn v e) =
  let (v', z0, nn0) = liftVecV tm nn  v
      (e', z1, nn1) = liftVecE tm nn0 e
  in  (Assn v' e', z0++z1, nn1)
liftVecS tm nn (Incr v e) =
  let (v', z0, nn0) = liftVecV tm nn  v
      (e', z1, nn1) = liftVecE tm nn0 e
  in  (Incr v' e', z0++z1, nn1)
-- todo: call sample and ldf; add dims
liftVecS tm nn (Call f r a) =
  let (a', z, nn0) = liftMany liftVecE tm nn a
  in  (Call f r a', z, nn0)
liftVecS tm nn (Comment s) = (Comment s, [], nn)

liftVecE tm nn (Var v) = let (v',z,nn') = liftVecV tm nn v in (Var v', z, nn')
liftVecE tm nn (Con c) = (Con c, [], nn)
liftVecE tm nn f0@(Fun f el) 
  | null (indices $ inferCodeType tm f0) = (Fun f el', z, nn1)
  | f `elem` ["vec","ID"] = (var, (var, inferCodeType tm f0, Just (Call "vec" [] (var:el'))) : z, nn1+1)
  where var = Var (V ("vec_var_" ++ show nn1) [])
        (el', z, nn1) = liftMany liftVecE tm nn el
liftVecE tm nn (Fun f [a,b])
  | f `elem` [".+",".-",".*","./"] = (Fun fn (var:a':b':dmm), [(var, t0, Nothing)] ++ aRest ++ bRest, nn2+1)
  | f `elem` [ "+", "-", "*", "/"] = (Fun fn (var:a':b':dmm), [(var, t0, Nothing)] ++ aRest ++ bRest, nn2+1)
  where t0@(Type t0t t0i) = inferCodeType tm b 
        var = Var (V ("vec_var_" ++ show nn2) [])
        (a', aRest, nn1) = liftVecE tm nn  a
        (b', bRest, nn2) = liftVecE tm nn1 b

        fn = (case last f of { '+' -> "add" ; '-' -> "sub" ; '*' -> "mult" ; '/' -> "div" }) ++ "_" ++
             (if head f == '.' then "const" else "vec") ++ "_" ++ (case t0t of { T0Int -> "i" ; _ -> "r" }) ++ "_" ++ 
             show (length t0i)

        (dmm,[]) = codifyMany (codifyExpr tm) $ concatMap rangeToExpr t0i
liftVecE tm nn f = error ("Code.liftVecE: cannot lift " ++ show f)

liftVecV tm nn (V id el) =
  let (el', z, nn0) = liftMany liftVecE tm nn el
  in  (V id el', z, nn0)


unliftVecLoops :: Function -> Function
unliftVecLoops f = f { body = uvl (map fst $ params f) [] $ body f}
  where
    -- note: all dvv loops will be top-level!
    uvl qual toadd [] = []
    uvl qual toadd (b@(Loop id lo hi body):bs)
      | "dvv_loop_var_" `isPrefixOf` id = -- && trace (show (b,getEVars lo ++ getEVars hi,qual,remainingVars)) True =
      if null remainingVars 
        then b : uvl qual toadd bs
        else     uvl qual ((remainingVars,b):toadd) bs
      where
        remainingVars = filter (not . (`elem` qual)) $ nub $ (getEVars lo ++ getEVars hi)
        getEVars e = [id | V id _ <- listify (const True) e]
    uvl qual toadd ((Loop id lo hi body):bs) = (Loop id lo hi (toadd_now ++ body')) : uvl qual toadd bs
      where 
        toadd0    = [(vl \\ [id], b) | (vl,b) <- toadd]
        toadd_now = map snd $ filter (null.fst) toadd0
        toadd'    = filter (not.null.fst) toadd0
        body'     = uvl (id:qual) toadd' body
    uvl qual toadd (b:bs) = b : uvl qual toadd bs
